@@ -4,33 +4,35 @@ These classes are the much bigger part of the whole API, because they contain
 the interesting information about completion and goto operations.
 """
 import warnings
-from itertools import chain
 import re
 
-from jedi._compatibility import unicode, use_metaclass
+from jedi._compatibility import u
 from jedi import settings
 from jedi import common
-from jedi.parser import tree
-from jedi.evaluate.cache import memoize_default, CachedMetaClass
+from jedi.parser.cache import parser_cache
+from jedi.cache import memoize_method
 from jedi.evaluate import representation as er
-from jedi.evaluate import iterable
+from jedi.evaluate import instance
 from jedi.evaluate import imports
 from jedi.evaluate import compiled
-from jedi.api import keywords
-from jedi.evaluate.finder import filter_definition_names
+from jedi.evaluate.filters import ParamName
+from jedi.api.keywords import KeywordName
 
 
-def defined_names(evaluator, scope):
+def _sort_names_by_start_pos(names):
+    return sorted(names, key=lambda s: s.start_pos or (0, 0))
+
+
+def defined_names(evaluator, context):
     """
     List sub-definitions (e.g., methods in class).
 
     :type scope: Scope
     :rtype: list of Definition
     """
-    dct = scope.names_dict
-    names = list(chain.from_iterable(dct.values()))
-    names = filter_definition_names(names, scope)
-    return [Definition(evaluator, d) for d in sorted(names, key=lambda s: s.start_pos)]
+    filter = next(context.get_filters(search_global=True))
+    names = [name for name in filter.values()]
+    return [Definition(evaluator, n) for n in _sort_names_by_start_pos(names)]
 
 
 class BaseDefinition(object):
@@ -51,8 +53,6 @@ class BaseDefinition(object):
 
     _tuple_mapping = dict((tuple(k.split('.')), v) for (k, v) in {
         'argparse._ActionsContainer': 'argparse.ArgumentParser',
-        '_sre.SRE_Match': 're.MatchObject',
-        '_sre.SRE_Pattern': 're.RegexObject',
     }.items())
 
     def __init__(self, evaluator, name):
@@ -61,15 +61,14 @@ class BaseDefinition(object):
         """
         An instance of :class:`jedi.parser.reprsentation.Name` subclass.
         """
-        self._definition = evaluator.wrap(self._name.get_definition())
-        self.is_keyword = isinstance(self._definition, keywords.Keyword)
+        self.is_keyword = isinstance(self._name, KeywordName)
 
         # generate a path to the definition
-        self._module = name.get_parent_until()
+        self._module = name.get_root_context()
         if self.in_builtin_module():
             self.module_path = None
         else:
-            self.module_path = self._module.path
+            self.module_path = self._module.py__file__()
             """Shows the file path of a module. e.g. ``/usr/lib/python2.7/os.py``"""
 
     @property
@@ -81,17 +80,7 @@ class BaseDefinition(object):
 
         :rtype: str or None
         """
-        return unicode(self._name)
-
-    @property
-    def start_pos(self):
-        """
-        .. deprecated:: 0.7.0
-           Use :attr:`.line` and :attr:`.column` instead.
-        .. todo:: Remove!
-        """
-        warnings.warn("Use line/column instead.", DeprecationWarning)
-        return self._name.start_pos
+        return self._name.string_name
 
     @property
     def type(self):
@@ -130,7 +119,7 @@ class BaseDefinition(object):
         >>> defs = sorted(defs, key=lambda d: d.line)
         >>> defs                           # doctest: +NORMALIZE_WHITESPACE
         [<Definition module keyword>, <Definition class C>,
-         <Definition class D>, <Definition def f>]
+         <Definition instance D>, <Definition def f>]
 
         Finally, here is what you can get from :attr:`type`:
 
@@ -144,45 +133,51 @@ class BaseDefinition(object):
         'function'
 
         """
-        stripped = self._definition
-        if isinstance(stripped, er.InstanceElement):
-            stripped = stripped.var
+        tree_name = self._name.tree_name
+        resolve = False
+        if tree_name is not None:
+            # TODO move this to their respective names.
+            definition = tree_name.get_definition()
+            if definition.type == 'import_from' and \
+                    tree_name in definition.get_defined_names():
+                resolve = True
 
-        if isinstance(stripped, compiled.CompiledObject):
-            return stripped.api_type()
-        elif isinstance(stripped, iterable.Array):
-            return 'instance'
-        elif isinstance(stripped, tree.Import):
-            return 'import'
-
-        string = type(stripped).__name__.lower().replace('wrapper', '')
-        if string == 'exprstmt':
-            return 'statement'
-        else:
-            return string
+        if isinstance(self._name, imports.SubModuleName) or resolve:
+            for context in self._name.infer():
+                return context.api_type
+        return self._name.api_type
 
     def _path(self):
         """The path to a module/class/function definition."""
-        path = []
-        par = self._definition
-        while par is not None:
-            if isinstance(par, tree.Import):
-                path += imports.ImportWrapper(self._evaluator, self._name).import_path
-                break
-            try:
-                name = par.name
-            except AttributeError:
-                pass
+        def to_reverse():
+            name = self._name
+            if name.api_type == 'module':
+                try:
+                    name = list(name.infer())[0].name
+                except IndexError:
+                    pass
+
+            if name.api_type == 'module':
+                module_context, = name.infer()
+                for n in reversed(module_context.py__name__().split('.')):
+                    yield n
             else:
-                if isinstance(par, er.ModuleWrapper):
-                    # TODO just make the path dotted from the beginning, we
-                    # shouldn't really split here.
-                    path[0:0] = par.py__name__().split('.')
-                    break
+                yield name.string_name
+
+            parent_context = name.parent_context
+            while parent_context is not None:
+                try:
+                    method = parent_context.py__name__
+                except AttributeError:
+                    try:
+                        yield parent_context.name.string_name
+                    except AttributeError:
+                        pass
                 else:
-                    path.insert(0, unicode(name))
-            par = par.parent
-        return path
+                    for name in reversed(method().split('.')):
+                        yield name
+                parent_context = parent_context.parent_context
+        return reversed(list(to_reverse()))
 
     @property
     def module_name(self):
@@ -196,7 +191,7 @@ class BaseDefinition(object):
         >>> print(d.module_name)                       # doctest: +ELLIPSIS
         json
         """
-        return str(self._module.name)
+        return self._module.name.string_name
 
     def in_builtin_module(self):
         """Whether this is a builtin module."""
@@ -205,18 +200,20 @@ class BaseDefinition(object):
     @property
     def line(self):
         """The line where the definition occurs (starting with 1)."""
-        if self.in_builtin_module():
+        start_pos = self._name.start_pos
+        if start_pos is None:
             return None
-        return self._name.start_pos[0]
+        return start_pos[0]
 
     @property
     def column(self):
         """The column where the definition occurs (starting with 0)."""
-        if self.in_builtin_module():
+        start_pos = self._name.start_pos
+        if start_pos is None:
             return None
-        return self._name.start_pos[1]
+        return start_pos[1]
 
-    def docstring(self, raw=False):
+    def docstring(self, raw=False, fast=True):
         r"""
         Return a document string for this completion object.
 
@@ -241,11 +238,16 @@ class BaseDefinition(object):
         >>> print(script.goto_definitions()[0].docstring(raw=True))
         Document for function f.
 
+        :param fast: Don't follow imports that are only one level deep like
+            ``import foo``, but follow ``from foo import bar``. This makes
+            sense for speed reasons. Completing `import a` is slow if you use
+            the ``foo.docstring(fast=False)`` on every object, because it
+            parses all libraries starting with ``a``.
         """
         if raw:
-            return _Help(self._definition).raw()
+            return _Help(self._name).raw(fast=fast)
         else:
-            return _Help(self._definition).full()
+            return _Help(self._name).full(fast=fast)
 
     @property
     def doc(self):
@@ -270,7 +272,7 @@ class BaseDefinition(object):
     @property
     def description(self):
         """A textual description of the object."""
-        return unicode(self._name)
+        return u(self._name.string_name)
 
     @property
     def full_name(self):
@@ -291,11 +293,12 @@ class BaseDefinition(object):
         >>> print(script.goto_definitions()[0].full_name)
         os.path.join
 
-        Notice that it correctly returns ``'os.path.join'`` instead of
-        (for example) ``'posixpath.join'``.
-
+        Notice that it returns ``'os.path.join'`` instead of (for example)
+        ``'posixpath.join'``. This is not correct, since the modules name would
+        be ``<module 'posixpath' ...>```. However most users find the latter
+        more practical.
         """
-        path = [unicode(p) for p in self._path()]
+        path = list(self._path())
         # TODO add further checks, the mapping should only occur on stdlib.
         if not path:
             return None  # for keywords the path is empty
@@ -309,89 +312,122 @@ class BaseDefinition(object):
         return '.'.join(path if path[0] else path[1:])
 
     def goto_assignments(self):
-        defs = self._evaluator.goto(self._name)
-        return [Definition(self._evaluator, d) for d in defs]
+        if self._name.tree_name is None:
+            return self
 
-    @memoize_default()
-    def _follow_statements_imports(self):
-        """
-        Follow both statements and imports, as far as possible.
-        """
-        if self._definition.isinstance(tree.ExprStmt):
-            return self._evaluator.eval_statement(self._definition)
-        elif self._definition.isinstance(tree.Import):
-            return imports.ImportWrapper(self._evaluator, self._name).follow()
-        else:
-            return [self._definition]
+        names = self._evaluator.goto(self._name.parent_context, self._name.tree_name)
+        return [Definition(self._evaluator, n) for n in names]
+
+    def _goto_definitions(self):
+        # TODO make this function public.
+        return [Definition(self._evaluator, d.name) for d in self._name.infer()]
 
     @property
-    @memoize_default()
+    @memoize_method
     def params(self):
         """
         Raises an ``AttributeError``if the definition is not callable.
         Otherwise returns a list of `Definition` that represents the params.
         """
-        followed = self._follow_statements_imports()
+        def get_param_names(context):
+            param_names = []
+            if context.api_type == 'function':
+                param_names = list(context.get_param_names())
+                if isinstance(context, instance.BoundMethod):
+                    param_names = param_names[1:]
+            elif isinstance(context, (instance.AbstractInstanceContext, er.ClassContext)):
+                if isinstance(context, er.ClassContext):
+                    search = '__init__'
+                else:
+                    search = '__call__'
+                names = context.get_function_slot_names(search)
+                if not names:
+                    return []
+
+                # Just take the first one here, not optimal, but currently
+                # there's no better solution.
+                inferred = names[0].infer()
+                param_names = get_param_names(next(iter(inferred)))
+                if isinstance(context, er.ClassContext):
+                    param_names = param_names[1:]
+                return param_names
+            elif isinstance(context, compiled.CompiledObject):
+                return list(context.get_param_names())
+            return param_names
+
+        followed = list(self._name.infer())
         if not followed or not hasattr(followed[0], 'py__call__'):
             raise AttributeError()
-        followed = followed[0]  # only check the first one.
+        context = followed[0]  # only check the first one.
 
-        if followed.type == 'funcdef':
-            if isinstance(followed, er.InstanceElement):
-                params = followed.params[1:]
-            else:
-                params = followed.params
-        elif followed.isinstance(er.compiled.CompiledObject):
-            params = followed.params
-        else:
-            try:
-                sub = followed.get_subscope_by_name('__init__')
-                params = sub.params[1:]  # ignore self
-            except KeyError:
-                return []
-        return [_Param(self._evaluator, p.name) for p in params]
+        return [_Param(self._evaluator, n) for n in get_param_names(context)]
 
     def parent(self):
-        scope = self._definition.get_parent_scope()
-        scope = self._evaluator.wrap(scope)
-        return Definition(self._evaluator, scope.name)
+        context = self._name.parent_context
+        if context is None:
+            return None
+
+        if isinstance(context, er.FunctionExecutionContext):
+            # TODO the function context should be a part of the function
+            # execution context.
+            context = er.FunctionContext(
+                self._evaluator, context.parent_context, context.tree_node)
+        return Definition(self._evaluator, context.name)
 
     def __repr__(self):
         return "<%s %s>" % (type(self).__name__, self.description)
 
+    def get_line_code(self, before=0, after=0):
+        """
+        Returns the line of code where this object was defined.
+
+        :param before: Add n lines before the current line to the output.
+        :param after: Add n lines after the current line to the output.
+
+        :return str: Returns the line(s) of code or an empty string if it's a
+                     builtin.
+        """
+        if self.in_builtin_module():
+            return ''
+
+        path = self._name.get_root_context().py__file__()
+        lines = parser_cache[path].lines
+
+        line_nr = self._name.start_pos[0]
+        start_line_nr = line_nr - before
+        return ''.join(lines[start_line_nr:line_nr + after + 1])
+
 
 class Completion(BaseDefinition):
     """
     `Completion` objects are returned from :meth:`api.Script.completions`. They
     provide additional information about a completion.
     """
-    def __init__(self, evaluator, name, needs_dot, like_name_length):
+    def __init__(self, evaluator, name, stack, like_name_length):
         super(Completion, self).__init__(evaluator, name)
 
-        self._needs_dot = needs_dot
         self._like_name_length = like_name_length
+        self._stack = stack
 
         # Completion objects with the same Completion name (which means
         # duplicate items in the completion)
         self._same_name_completions = []
 
     def _complete(self, like_name):
-        dot = '.' if self._needs_dot else ''
         append = ''
         if settings.add_bracket_after_function \
                 and self.type == 'Function':
             append = '('
 
-        if settings.add_dot_after_module:
-            if isinstance(self._definition, tree.Module):
-                append += '.'
-        if isinstance(self._definition, tree.Param):
-            append += '='
+        if isinstance(self._name, ParamName) and self._stack is not None:
+            node_names = list(self._stack.get_node_names(self._evaluator.grammar))
+            if 'trailer' in node_names and 'argument' not in node_names:
+                append += '='
 
-        name = str(self._name)
+        name = self._name.string_name
         if like_name:
             name = name[self._like_name_length:]
-        return dot + name + append
+        return name + append
 
     @property
     def complete(self):
@@ -402,90 +438,51 @@ class Completion(BaseDefinition):
 
         would return the string 'ce'. It also adds additional stuff, depending
         on your `settings.py`.
+
+        Assuming the following function definition::
+
+            def foo(param=0):
+                pass
+
+        completing ``foo(par`` would give a ``Completion`` which `complete`
+        would be `am=`
+
+
         """
         return self._complete(True)
 
     @property
     def name_with_symbols(self):
         """
-        Similar to :attr:`name`, but like :attr:`name`
-        returns also the symbols, for example::
+        Similar to :attr:`name`, but like :attr:`name` returns also the
+        symbols, for example assuming the following function definition::
 
-            list()
+            def foo(param=0):
+                pass
+
+        completing ``foo(`` would give a ``Completion`` which
+        ``name_with_symbols`` would be "param=".
 
-        would return ``.append`` and others (which means it adds a dot).
         """
         return self._complete(False)
 
+    def docstring(self, raw=False, fast=True):
+        if self._like_name_length >= 3:
+            # In this case we can just resolve the like name, because we
+            # wouldn't load like > 100 Python modules anymore.
+            fast = False
+        return super(Completion, self,).docstring(raw, fast)
+
     @property
     def description(self):
         """Provide a description of the completion object."""
-        if self._definition is None:
-            return ''
-        t = self.type
-        if t == 'statement' or t == 'import':
-            desc = self._definition.get_code()
-        else:
-            desc = '.'.join(unicode(p) for p in self._path())
-
-        line = '' if self.in_builtin_module else '@%s' % self.line
-        return '%s: %s%s' % (t, desc, line)
+        # TODO improve the class structure.
+        return Definition.description.__get__(self)
 
     def __repr__(self):
-        return '<%s: %s>' % (type(self).__name__, self._name)
+        return '<%s: %s>' % (type(self).__name__, self._name.string_name)
 
-    def docstring(self, raw=False, fast=True):
-        """
-        :param fast: Don't follow imports that are only one level deep like
-            ``import foo``, but follow ``from foo import bar``. This makes
-            sense for speed reasons. Completing `import a` is slow if you use
-            the ``foo.docstring(fast=False)`` on every object, because it
-            parses all libraries starting with ``a``.
-        """
-        definition = self._definition
-        if isinstance(definition, tree.Import):
-            i = imports.ImportWrapper(self._evaluator, self._name)
-            if len(i.import_path) > 1 or not fast:
-                followed = self._follow_statements_imports()
-                if followed:
-                    # TODO: Use all of the followed objects as input to Documentation.
-                    definition = followed[0]
-
-        if raw:
-            return _Help(definition).raw()
-        else:
-            return _Help(definition).full()
-
-    @property
-    def type(self):
-        """
-        The type of the completion objects. Follows imports. For a further
-        description, look at :attr:`jedi.api.classes.BaseDefinition.type`.
-        """
-        if isinstance(self._definition, tree.Import):
-            i = imports.ImportWrapper(self._evaluator, self._name)
-            if len(i.import_path) <= 1:
-                return 'module'
-
-            followed = self.follow_definition()
-            if followed:
-                # Caveat: Only follows the first one, ignore the other ones.
-                # This is ok, since people are almost never interested in
-                # variations.
-                return followed[0].type
-        return super(Completion, self).type
-
-    @memoize_default()
-    def _follow_statements_imports(self):
-        # imports completion is very complicated and needs to be treated
-        # separately in Completion.
-        definition = self._definition
-        if definition.isinstance(tree.Import):
-            i = imports.ImportWrapper(self._evaluator, self._name)
-            return i.follow()
-        return super(Completion, self)._follow_statements_imports()
-
-    @memoize_default()
+    @memoize_method
     def follow_definition(self):
         """
         Return the original definitions. I strongly recommend not using it for
@@ -495,11 +492,11 @@ class Completion(BaseDefinition):
         follows all results. This means with 1000 completions (e.g.  numpy),
         it's just PITA-slow.
         """
-        defs = self._follow_statements_imports()
+        defs = self._name.infer()
         return [Definition(self._evaluator, d.name) for d in defs]
 
 
-class Definition(use_metaclass(CachedMetaClass, BaseDefinition)):
+class Definition(BaseDefinition):
     """
     *Definition* objects are returned from :meth:`api.Script.goto_assignments`
     or :meth:`api.Script.goto_definitions`.
@@ -535,45 +532,25 @@ class Definition(use_metaclass(CachedMetaClass, BaseDefinition)):
         'class C'
 
         """
-        d = self._definition
-        if isinstance(d, er.InstanceElement):
-            d = d.var
+        typ = self.type
+        tree_name = self._name.tree_name
+        if typ in ('function', 'class', 'module', 'instance') or tree_name is None:
+            if typ == 'function':
+                # For the description we want a short and a pythonic way.
+                typ = 'def'
+            return typ + ' ' + u(self._name.string_name)
+        elif typ == 'param':
+            return typ + ' ' + tree_name.get_definition().get_description()
 
-        if isinstance(d, compiled.CompiledObject):
-            typ = d.api_type()
-            if typ == 'instance':
-                typ = 'class'  # The description should be similar to Py objects.
-            d = typ + ' ' + d.name.get_code()
-        elif isinstance(d, iterable.Array):
-            d = 'class ' + d.type
-        elif isinstance(d, (tree.Class, er.Class, er.Instance)):
-            d = 'class ' + unicode(d.name)
-        elif isinstance(d, (er.Function, tree.Function)):
-            d = 'def ' + unicode(d.name)
-        elif isinstance(d, tree.Module):
-            # only show module name
-            d = 'module %s' % self.module_name
-        elif isinstance(d, tree.Param):
-            d = d.get_code().strip()
-            if d.endswith(','):
-                d = d[:-1]  # Remove the comma.
-        else:  # ExprStmt
-            try:
-                first_leaf = d.first_leaf()
-            except AttributeError:
-                # `d` is already a Leaf (Name).
-                first_leaf = d
-            # Remove the prefix, because that's not what we want for get_code
-            # here.
-            old, first_leaf.prefix = first_leaf.prefix, ''
-            try:
-                d = d.get_code()
-            finally:
-                first_leaf.prefix = old
+        definition = tree_name.get_definition()
+        # Remove the prefix, because that's not what we want for get_code
+        # here.
+        txt = definition.get_code(include_prefix=False)
         # Delete comments:
-        d = re.sub('#[^\n]+\n', ' ', d)
+        txt = re.sub('#[^\n]+\n', ' ', txt)
         # Delete multi spaces/newlines
-        return re.sub('\s+', ' ', d).strip()
+        txt = re.sub('\s+', ' ', txt).strip()
+        return txt
 
     @property
     def desc_with_module(self):
@@ -589,26 +566,28 @@ class Definition(use_metaclass(CachedMetaClass, BaseDefinition)):
         position = '' if self.in_builtin_module else '@%s' % (self.line)
         return "%s:%s%s" % (self.module_name, self.description, position)
 
-    @memoize_default()
+    @memoize_method
     def defined_names(self):
         """
         List sub-definitions (e.g., methods in class).
 
         :rtype: list of Definition
         """
-        defs = self._follow_statements_imports()
-        # For now we don't want base classes or evaluate decorators.
-        defs = [d.base if isinstance(d, (er.Class, er.Function)) else d for d in defs]
-        iterable = (defined_names(self._evaluator, d) for d in defs)
-        iterable = list(iterable)
-        return list(chain.from_iterable(iterable))
+        defs = self._name.infer()
+        return sorted(
+            common.unite(defined_names(self._evaluator, d) for d in defs),
+            key=lambda s: s._name.start_pos or (0, 0)
+        )
 
     def is_definition(self):
         """
         Returns True, if defined as a name in a statement, function or class.
         Returns False, if it's a reference to such a definition.
         """
-        return self._name.is_definition()
+        if self._name.tree_name is None:
+            return True
+        else:
+            return self._name.tree_name.is_definition()
 
     def __eq__(self, other):
         return self._name.start_pos == other._name.start_pos \
@@ -629,11 +608,11 @@ class CallSignature(Definition):
     It knows what functions you are currently in. e.g. `isinstance(` would
     return the `isinstance` function. without `(` it would return nothing.
     """
-    def __init__(self, evaluator, executable_name, call_stmt, index, key_name):
+    def __init__(self, evaluator, executable_name, bracket_start_pos, index, key_name_str):
         super(CallSignature, self).__init__(evaluator, executable_name)
         self._index = index
-        self._key_name = key_name
-        self._call_stmt = call_stmt
+        self._key_name_str = key_name_str
+        self._bracket_start_pos = bracket_start_pos
 
     @property
     def index(self):
@@ -641,21 +620,24 @@ class CallSignature(Definition):
         The Param index of the current call.
         Returns None if the index cannot be found in the curent call.
         """
-        if self._key_name is not None:
+        if self._key_name_str is not None:
             for i, param in enumerate(self.params):
-                if self._key_name == param.name:
+                if self._key_name_str == param.name:
                     return i
-            if self.params and self.params[-1]._name.get_definition().stars == 2:
-                return i
-            else:
-                return None
+            if self.params:
+                param_name = self.params[-1]._name
+                if param_name.tree_name is not None:
+                    if param_name.tree_name.get_definition().stars == 2:
+                        return i
+            return None
 
         if self._index >= len(self.params):
-
             for i, param in enumerate(self.params):
-                # *args case
-                if param._name.get_definition().stars == 1:
-                    return i
+                tree_name = param._name.tree_name
+                if tree_name is not None:
+                    # *args case
+                    if tree_name.get_definition().stars == 1:
+                        return i
             return None
         return self._index
 
@@ -665,7 +647,7 @@ class CallSignature(Definition):
         The indent of the bracket that is responsible for the last function
         call.
         """
-        return self._call_stmt.end_pos
+        return self._bracket_start_pos
 
     @property
     def call_name(self):
@@ -677,7 +659,7 @@ class CallSignature(Definition):
         The name (e.g. 'isinstance') as a string.
         """
         warnings.warn("Use name instead.", DeprecationWarning)
-        return unicode(self.name)
+        return self.name
 
     @property
     def module(self):
@@ -686,11 +668,11 @@ class CallSignature(Definition):
            Use :attr:`.module_name` for the module name.
         .. todo:: Remove!
         """
-        return self._executable.get_parent_until()
+        return self._executable.get_root_node()
 
     def __repr__(self):
-        return '<%s: %s index %s>' % (type(self).__name__, self._name,
-                                      self.index)
+        return '<%s: %s index %s>' % \
+            (type(self).__name__, self._name.string_name, self.index)
 
 
 class _Param(Definition):
@@ -717,19 +699,42 @@ class _Help(object):
     def __init__(self, definition):
         self._name = definition
 
-    def full(self):
-        try:
-            return self._name.doc
-        except AttributeError:
-            return self.raw()
+    @memoize_method
+    def _get_node(self, fast):
+        if isinstance(self._name, (compiled.CompiledContextName, compiled.CompiledName)):
+            followed = self._name.infer()
+            if followed:
+                return next(iter(followed))
+            return None
 
-    def raw(self):
+        if self._name.api_type == 'module' and not fast:
+            followed = self._name.infer()
+            if followed:
+                # TODO: Use all of the followed objects as input to Documentation.
+                context = next(iter(followed))
+                return context.tree_node
+        if self._name.tree_name is None:
+            return None
+        return self._name.tree_name.get_definition()
+
+    def full(self, fast=True):
+        node = self._get_node(fast)
+        try:
+            return node.doc
+        except AttributeError:
+            return self.raw(fast)
+
+    def raw(self, fast=True):
         """
         The raw docstring ``__doc__`` for any object.
 
         See :attr:`doc` for example.
         """
+        node = self._get_node(fast)
+        if node is None:
+            return ''
+
         try:
-            return self._name.raw_doc
+            return node.raw_doc
         except AttributeError:
             return ''
